@@ -3,22 +3,28 @@ from datetime import datetime
 from pathlib import Path
 from typing import Optional, Tuple
 
-from fastapi import APIRouter, Depends, WebSocketDisconnect
+from fastapi import (
+    APIRouter,
+    Depends,
+    HTTPException,
+    Request,
+    Response,
+    WebSocketDisconnect,
+    status,
+)
+from fastapi.responses import PlainTextResponse
 from fps.hooks import register_router  # type: ignore
-from fps_contents.routes import read_content, write_content  # type: ignore
-
-try:
-    from fps_contents.watchfiles import awatch
-
-    has_awatch = True
-except ImportError:
-    has_awatch = False
 from fps_auth_base import websocket_auth  # type: ignore
+from fps_auth_base import User, current_user
+from fps_contents.fileid import FileIdManager
+from fps_contents.routes import read_content, write_content  # type: ignore
 from jupyter_ydoc import ydocs as YDOCS  # type: ignore
 from ypy_websocket.websocket_server import WebsocketServer, YRoom  # type: ignore
 from ypy_websocket.ystore import BaseYStore, SQLiteYStore, YDocNotFound  # type: ignore
 from ypy_websocket.yutils import YMessageType  # type: ignore
 
+from .models import CreateRoomId
+
 YFILE = YDOCS["file"]
 AWARENESS = 1
 RENAME_SESSION = 127
@@ -35,6 +41,12 @@ def to_datetime(iso_date: str) -> datetime:
     return datetime.fromisoformat(iso_date.rstrip("Z"))
 
 
+@router.on_event("startup")
+async def startup():
+    # start indexing in the background
+    FileIdManager()
+
+
 @router.websocket("/api/yjs/{path:path}")
 async def websocket_endpoint(
     path,
@@ -106,7 +118,7 @@ class JupyterWebsocketServer(WebsocketServer):
                 # it is a stored document (e.g. a notebook)
                 file_format, file_type, file_path = path.split(":", 2)
                 p = Path(file_path)
-                updates_file_path = str(p.parent / f".{file_type}:{p.name}.y")
+                updates_file_path = (p.parent / f".{file_type}:{p.name}.y").as_posix()
                 ystore = JupyterSQLiteYStore(path=updates_file_path)  # FIXME: pass in config
                 self.rooms[path] = DocumentRoom(file_type, ystore)
             else:
@@ -123,12 +135,18 @@ class YDocWebSocketHandler:
     def __init__(self, websocket, path, permissions):
         self.websocket = websocket
         self.can_write = permissions is None or "write" in permissions.get("yjs", [])
+        self.lock = asyncio.Lock()
         self.room = self.websocket_server.get_room(self.websocket.path)
         self.set_file_info(path)
 
-    def get_file_info(self) -> Tuple[str, str, str]:
+    async def get_file_info(self) -> Tuple[str, str, str]:
         room_name = self.websocket_server.get_room_name(self.room)
-        file_format, file_type, file_path = room_name.split(":", 2)
+        file_format, file_type, file_id = room_name.split(":", 2)
+        file_path = await FileIdManager().get_path(file_id)
+        if file_path is None:
+            raise RuntimeError(f"File {self.room.document.path} cannot be found anymore")
+        if file_path != self.room.document.path:
+            self.room.document.path = file_path
         return file_format, file_type, file_path
 
     def set_file_info(self, value: str) -> None:
@@ -145,7 +163,7 @@ class YDocWebSocketHandler:
             self.room.cleaner.cancel()
 
         if not self.room.is_transient and not self.room.ready:
-            file_format, file_type, file_path = self.get_file_info()
+            file_format, file_type, file_path = await self.get_file_info()
             is_notebook = file_type == "notebook"
             model = await read_content(file_path, True, as_json=is_notebook)
             self.last_modified = to_datetime(model.last_modified)
@@ -174,7 +192,7 @@ class YDocWebSocketHandler:
                 self.room.document.observe(self.on_document_change)
 
         await self.websocket_server.serve(self.websocket)
-        if not self.room.is_transient and self.room.clients == [self.websocket]:
+        if not self.room.is_transient and not self.room.clients:
             # no client in this room after we disconnect
             # keep the document for a while in case someone reconnects
             self.room.cleaner = asyncio.create_task(self.clean_room())
@@ -212,23 +230,22 @@ class YDocWebSocketHandler:
         return skip
 
     async def watch_file(self):
-        if has_awatch:
-            file_format, file_type, file_path = self.get_file_info()
-            async for changes in awatch(file_path):
-                await self.maybe_load_document()
-        else:
-            # contents plugin doesn't provide watcher, fall back to polling
-            poll_interval = 1  # FIXME: pass in config
-            if not poll_interval:
-                self.room.watcher = None
-                return
-            while True:
-                await asyncio.sleep(poll_interval)
+        file_format, file_type, file_path = await self.get_file_info()
+        while True:
+            watcher = FileIdManager().watch(file_path)
+            async for changes in watcher:
+                file_format, file_type, new_file_path = await self.get_file_info()
+                if new_file_path != file_path:
+                    # file was renamed
+                    FileIdManager().unwatch(file_path, watcher)
+                    file_path = new_file_path
+                    break
                 await self.maybe_load_document()
 
     async def maybe_load_document(self):
-        file_format, file_type, file_path = self.get_file_info()
-        model = await read_content(file_path, False)
+        file_format, file_type, file_path = await self.get_file_info()
+        async with self.lock:
+            model = await read_content(file_path, False)
         # do nothing if the file was saved by us
         if self.last_modified < to_datetime(model.last_modified):
             is_notebook = file_type == "notebook"
@@ -266,11 +283,12 @@ class YDocWebSocketHandler:
         await asyncio.sleep(1)
         # if the room cannot be found, don't save
         try:
-            file_format, file_type, file_path = self.get_file_info()
+            file_format, file_type, file_path = await self.get_file_info()
         except Exception:
             return
         is_notebook = file_type == "notebook"
-        model = await read_content(file_path, True, as_json=is_notebook)
+        async with self.lock:
+            model = await read_content(file_path, True, as_json=is_notebook)
         if self.last_modified < to_datetime(model.last_modified):
             # file changed on disk, let's revert
             self.room.document.source = model.content
@@ -287,10 +305,34 @@ class YDocWebSocketHandler:
                 "path": file_path,
                 "type": file_type,
             }
-            await write_content(content)
-            model = await read_content(file_path, False)
-            self.last_modified = to_datetime(model.last_modified)
+            async with self.lock:
+                await write_content(content)
+                model = await read_content(file_path, False)
+                self.last_modified = to_datetime(model.last_modified)
         self.room.document.dirty = False
 
 
+@router.put("/api/yjs/roomid/{path:path}", status_code=200, response_class=PlainTextResponse)
+async def create_roomid(
+    path,
+    request: Request,
+    response: Response,
+    user: User = Depends(current_user(permissions={"contents": ["read"]})),
+):
+    # we need to process the request manually
+    # see https://github.com/tiangolo/fastapi/issues/3373#issuecomment-1306003451
+    create_room_id = CreateRoomId(**(await request.json()))
+    ws_url = f"{create_room_id.format}:{create_room_id.type}:"
+    idx = await FileIdManager().get_id(path)
+    if idx is not None:
+        return ws_url + idx
+
+    idx = await FileIdManager().index(path)
+    if idx is None:
+        raise HTTPException(status_code=404, detail=f"File {path} does not exist")
+
+    response.status_code = status.HTTP_201_CREATED
+    return ws_url + idx
+
+
 r = register_router(router)
